/*
 * dingdictionary.cpp
 *
 * Copyright (c) 2005 Michael Rex <me@rexi.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "dingdictionary.h"
#include <kstandarddirs.h>
#include <kdebug.h>
#include <kglobal.h>
#include <klocale.h>
#include <qregexp.h>
#include <qvaluelist.h>
#include <qprocess.h>
#include <qfile.h>

DingDictionary::DingDictionary(QObject* parent, const char* name) : QObject(parent, name), PATH(locate("data", "kding/de-en.txt")), VERSION(determineDictionaryVersion()), proc(0), reslist(0) {
    
}

DingDictionary::~DingDictionary() {
    
}

void DingDictionary::search(const QString& text) {
    // if there is already another search process running emit a signal and
    // return without doing anything
    if(proc != 0) {
        emit searchRunning();
        return;
    }
    
    // do some preparatory tasks:
    // we need to store the search term to properly sort the entries in the
    // result list by their priority ...
    search_term = text;
    // ... and we delete the old result list as it will be replaced by whatever
    // this search returns
    delete reslist;
    reslist = 0;
    
    // now construct the QProcess object and finally start searching
    proc = new QProcess();
    proc->addArgument("egrep");
    proc->addArgument("-i");    // case insensitive
    proc->addArgument("-w");    // match only whole words
    //proc->addArgument("^[^#].*\\b" + text);
    // this check is needed to pass a correctly encoded search term to the
    // grep command
    if(QString("UTF-8") == KGlobal::locale()->encoding()) {
        proc->addArgument(text);
    } else {
        proc->addArgument(text.utf8());
    }
    proc->addArgument(PATH);
    
    connect(proc, SIGNAL(processExited()), this, SLOT(readDataFromProcess()));
    
    kdDebug() << proc->arguments() << endl;
    proc->start();
}

void DingDictionary::readDataFromProcess() {
    kdDebug() << "egrep exit status: " << proc->exitStatus() << endl;
    
    reslist = new ResultList();
    while(proc->canReadLineStdout()) {
        reslist->append(ResultItem(QString::fromUtf8(proc->readLineStdout())));
    }
    
    // we don't need this anymore, so delete the object and set the variable to
    // 0 so the check for an already running search process works
    delete proc;
    proc = 0;
    
    sortByPriority(reslist);
    
    emit searchFinished();
}

const ResultList& DingDictionary::results() const {
    return *reslist;
}

void DingDictionary::sortByPriority(ResultList* list) {
    const int VALUE_ANFANG = 100;
    const int VALUE_ERSTEZEILE = 100;
    const int VALUE_EINZELWORT = 80;
    const int VALUE_ABK = 10;
    
    QRegExp ml("\\|");
    QRegExp abkuerzung(" : ");
    QRegExp anfang_ger("^" + search_term + "(;| ::)", false);
    QRegExp anfang_eng(":: (to )?" + search_term + "(;|$)", false);
    QRegExp anfang_ger_ml("^" + search_term + "(;| ).*\\|.* ::", false);
    QRegExp anfang_eng_ml(":: (to )?" + search_term + "(;.*| )\\|", false);
    QRegExp einzelwort("; " + search_term + "(;|$)", false);
    QRegExp erstezeile_ger("^[^\\|]*" + search_term + ".*\\|.* ::", false);
    QRegExp erstezeile_eng(":: [^\\|]*" + search_term + ".*\\|", false);
    
    QRegExp brackets(" \\([^.\\)]*\\)");
    QRegExp curlybrackets(" \\{[^.\\}]*\\}");
    QRegExp squarebrackets(" \\[[^\\]]*\\]");
    
    QValueListIterator<ResultItem> end = list->end();
    for(QValueListIterator<ResultItem> it = list->begin(); it != end; ++it) {
        QString text = (*it).text();
        text.replace(brackets, "");
        text.replace(curlybrackets, "");
        text.replace(squarebrackets, "");
        
        if(ml.search(text) != -1) { // multiline result
            if((anfang_ger_ml.search(text) != -1) || (anfang_eng_ml.search(text) != -1)) {
                (*it).addToPriority(VALUE_ANFANG);
            } else if((erstezeile_ger.search(text) != -1) || (erstezeile_eng.search(text) != -1)) {
                (*it).addToPriority(VALUE_ERSTEZEILE);
            }
        } else { // singleline result
            if((anfang_ger.search(text) != -1) || (anfang_eng.search(text) != -1)) {
                (*it).addToPriority(VALUE_ANFANG);
            } else if(einzelwort.search(text) != -1) {
                (*it).addToPriority(VALUE_EINZELWORT);
            }
            
            if(abkuerzung.search(text) != -1) {
                (*it).addToPriority(VALUE_ABK);
            }
        }
    }
    
    qHeapSort(*list);
//    for(QValueListIterator<ResultItem> it = list->begin(); it != list->end(); ++it)
//        kdDebug() << (*it).priority() << " :: " << (*it).text() << endl;
}

const QString& DingDictionary::searchTerm() const {
    return search_term;
}

const QString& DingDictionary::dictionaryVersion() const {
    return VERSION;
}

QString DingDictionary::determineDictionaryVersion() {
    /*
     * The version of the dictionary is given in the first line of the file.
     * As of v1.5 it is of the form
     *   # Version :: 1.5 2007-04-09
     * for release versions or
     *   # Version :: devel 2007-06-10
     * for development versions of the dictionary.
     */
    
    QString version(i18n("dictionary version", "unknown"));
    
    QFile dictFile(PATH);
    if(dictFile.open(IO_ReadOnly)) {
        QTextStream stream(&dictFile);
        QString line;
        
        line = stream.readLine();
        if(line != 0) {
            int pos = line.find("::");
            if(pos != -1) {
                version = line.mid(pos + 2).stripWhiteSpace();
            }
        }
        
        dictFile.close();
    }
    
    return version;
}
